/*
 * Copyright 2017 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.pm.ActivityInfo;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;
import android.net.Uri;
import android.os.Build.VERSION;
import android.os.Build.VERSION_CODES;
import android.util.Log;
import androidx.annotation.Nullable;
import androidx.annotation.VisibleForTesting;
import com.google.android.apps.work.dpcsupport.AccountManagementWhitelistEnforcer.ContinuousEnforcer;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/** Helper class for {@link AccountManagementWhitelistEnforcer}. */
class AccountManagementWhitelistHelper {
  @VisibleForTesting
  static final String PREFERENCE_FILE =
      "com.google.android.apps.work.dpcsupport.AccountManagementWhitelistEnforcer";

  @VisibleForTesting static final String PREFERENCE_KEY_ADMIN_COMPONENT = "key_admin_component";

  @VisibleForTesting static final String PREFERENCE_KEY_PACKAGE_WHITELIST = "key_package_whitelist";

  @VisibleForTesting
  static final String PREFERENCE_KEY_AUTHENTICATOR_WHITELIST = "key_account_management_whitelist";

  private final Context context;

  AccountManagementWhitelistHelper(Context context) {
    this.context = context;
  }

  /**
   * Returns a {@link Runnable} which updates the package whitelist in shared preferences and
   * enforces it right away.
   */
  Runnable getUpdateAndEnforcePackageWhitelistRunnable(
      final ComponentName adminComponent,
      @Nullable final Set<String> packageWhitelist,
      final AuthenticatorTypesGetter authenticatorTypesGetter) {
    return new Runnable() {
      @Override
      public void run() {
        if (packageWhitelist != null && !assertContinuousEnforcerEnabled()) {
          return;
        }
        updatePackageWhitelist(adminComponent, packageWhitelist);
        enforce(
            adminComponent,
            packageWhitelist,
            getAuthenticatorWhitelist(),
            authenticatorTypesGetter.getAuthenticatorTypeToPackageNameMap(context));
      }
    };
  }

  /**
   * Returns a {@link Runnable} which updates the authenticator whitelist in shared preferences and
   * enforces it right away.
   */
  Runnable getUpdateAndEnforceAuthenticatorWhitelistRunnable(
      final ComponentName adminComponent,
      @Nullable final Set<String> authenticatorWhitelist,
      final AuthenticatorTypesGetter authenticatorTypesGetter) {
    return new Runnable() {
      @Override
      public void run() {
        if (authenticatorWhitelist != null && !assertContinuousEnforcerEnabled()) {
          return;
        }
        updateAuthenticatorWhitelist(adminComponent, authenticatorWhitelist);
        enforce(
            adminComponent,
            getPackageWhitelist(),
            authenticatorWhitelist,
            authenticatorTypesGetter.getAuthenticatorTypeToPackageNameMap(context));
      }
    };
  }

  private boolean assertContinuousEnforcerEnabled() {
    if (VERSION.SDK_INT >= VERSION_CODES.O) {
      if (!ContinuousEnforcer.hasRegistered) {
        Log.e(
            Constants.TAG,
            String.format(
                "Please register %s with the DeviceAdminService",
                ContinuousEnforcer.class.getName()));
        return false;
      }
    } else if (!isContinuousEnforcerDeclaredInManifest()) {
      Log.e(
          Constants.TAG,
          String.format(
              "Please ensure that %s has been declared in the Manifest.",
              ContinuousEnforcer.class.getName()));
      return false;
    }
    return true;
  }

  /**
   * Returns a {@link Runnable} which enforces the account management whitelist stored in shared
   * preferences by calling {@link DevicePolicyManager#setAccountManagementDisabled(ComponentName,
   * String, boolean)} on any authenticator account type on the profile or device that is not
   * included in the whitelist.
   */
  Runnable getEnforceRunnable(final AuthenticatorTypesGetter authenticatorTypesGetter) {
    return new Runnable() {
      @Override
      public void run() {
        DevicePolicyManagerHelper devicePolicyManagerHelper =
            new DevicePolicyManagerHelper(context);
        if (!devicePolicyManagerHelper.isDeviceOrProfileOwner()) {
          Log.e(
              Constants.TAG,
              "Unable to enforce account management whitelist: caller is not device/profile owner");
          return;
        }

        ComponentName adminComponent = getAdminComponent();
        Set<String> authenticatorWhitelist = getAuthenticatorWhitelist();
        Set<String> packageWhitelist = getPackageWhitelist();
        if (adminComponent == null
            || (packageWhitelist == null && authenticatorWhitelist == null)) {
          Log.w(Constants.TAG, "Account management whitelist has not been set up");
          return;
        }
        enforce(
            adminComponent,
            packageWhitelist,
            authenticatorWhitelist,
            authenticatorTypesGetter.getAuthenticatorTypeToPackageNameMap(context));
      }
    };
  }

  private void enforce(
      ComponentName adminComponent,
      @Nullable Set<String> packageWhitelist,
      @Nullable Set<String> authenticatorWhitelist,
      Map<String, String> authenticatorTypeToPackageNameMap) {
    DevicePolicyManager dpm =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);

    for (Entry<String, String> entry : authenticatorTypeToPackageNameMap.entrySet()) {
      String authenticatorType = entry.getKey();
      boolean isOnWhitelist =
          (packageWhitelist == null && authenticatorWhitelist == null)
              || (packageWhitelist != null && packageWhitelist.contains(entry.getValue()))
              || (authenticatorWhitelist != null
                  && authenticatorWhitelist.contains(authenticatorType));

      dpm.setAccountManagementDisabled(
          adminComponent, authenticatorType, /* disabled= */ !isOnWhitelist);
      Log.i(
          Constants.TAG,
          String.format(
              "account management for type %s: %s",
              authenticatorType, isOnWhitelist ? "ENABLED" : "DISABLED"));
    }
  }

  private void updatePackageWhitelist(
      ComponentName adminComponent, @Nullable Set<String> packageWhitelist) {
    context
        .getSharedPreferences(PREFERENCE_FILE, Context.MODE_PRIVATE)
        .edit()
        .putString(PREFERENCE_KEY_ADMIN_COMPONENT, adminComponent.flattenToString())
        .putStringSet(PREFERENCE_KEY_PACKAGE_WHITELIST, packageWhitelist)
        .commit();
  }

  private void updateAuthenticatorWhitelist(
      ComponentName adminComponent, @Nullable Set<String> authenticatorWhitelist) {
    context
        .getSharedPreferences(PREFERENCE_FILE, Context.MODE_PRIVATE)
        .edit()
        .putString(PREFERENCE_KEY_ADMIN_COMPONENT, adminComponent.flattenToString())
        .putStringSet(PREFERENCE_KEY_AUTHENTICATOR_WHITELIST, authenticatorWhitelist)
        .commit();
  }

  /**
   * Returns whether {@link ContinuousEnforcer} broadcast receiver is declared in the manifest. On
   * pre-Oreo devices this can be used to determine if the ContinuousEnforcer has been set up
   * properly during whitelist update.
   */
  private boolean isContinuousEnforcerDeclaredInManifest() {
    PackageManager pm = context.getApplicationContext().getPackageManager();
    Intent intent = new Intent(Intent.ACTION_PACKAGE_ADDED);
    intent.setData(Uri.fromParts("package", /* ssp= */ "", /* fragment= */ null));
    List<ResolveInfo> resolveInfos = pm.queryBroadcastReceivers(intent, /* flags= */  0);
    if (resolveInfos != null) {
      for (ResolveInfo resolveInfo : resolveInfos) {
        ActivityInfo receiverInfo = resolveInfo.activityInfo;
        if (receiverInfo.isEnabled()
            && receiverInfo.name.equals(ContinuousEnforcer.class.getName())) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Returns the active admin component which is a profile or device owner as stored in the shared
   * preferences, or null if such value is not present.
   */
  @Nullable
  private ComponentName getAdminComponent() {
    String flattenedAdmin =
        context
            .getSharedPreferences(PREFERENCE_FILE, Context.MODE_PRIVATE)
            .getString(PREFERENCE_KEY_ADMIN_COMPONENT, /* defValues= */ null);
    return flattenedAdmin == null ? null : ComponentName.unflattenFromString(flattenedAdmin);
  }

  /**
   * Returns a set of package names for which account management will be enabled as stored in the
   * shared preferences, or null if such value is not present.
   */
  @Nullable
  private Set<String> getPackageWhitelist() {
    return context
        .getSharedPreferences(PREFERENCE_FILE, Context.MODE_PRIVATE)
        .getStringSet(PREFERENCE_KEY_PACKAGE_WHITELIST, /* defValues= */ null);
  }

  /**
   * Returns a set of account types for which account management will be enabled as stored in the
   * shared preferences, or null if such value is not present.
   */
  @Nullable
  private Set<String> getAuthenticatorWhitelist() {
    return context
        .getSharedPreferences(PREFERENCE_FILE, Context.MODE_PRIVATE)
        .getStringSet(PREFERENCE_KEY_AUTHENTICATOR_WHITELIST, /* defValues= */ null);
  }
}
