/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import static com.google.android.apps.work.dpcsupport.PlayServicesAndStoreConstants.PLAY_SERVICES_MIN_VERSION;
import static com.google.android.apps.work.dpcsupport.PlayServicesAndStoreConstants.PLAY_SERVICES_MIN_VERSION_READABLE;

import android.accounts.Account;
import android.content.ComponentName;
import android.content.Context;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.util.Log;
import androidx.annotation.NonNull;
import com.google.android.gms.common.GoogleApiAvailability;
import java.lang.Thread.UncaughtExceptionHandler;

/**
 * Provides support for Android for Work Accounts, including ensuring that the device is in a state
 * which can support adding these accounts.
 */
public class AndroidForWorkAccountSupport {

  private final Context context;
  private final ComponentName admin;
  private final int minPlayServicesVersion;

  public AndroidForWorkAccountSupport(Context context, ComponentName admin) {
    this(context, admin, GoogleApiAvailability.GOOGLE_PLAY_SERVICES_VERSION_CODE);
  }

  /**
   * @param minPlayServicesVersion the minimum Play Services version that should be enforced, by
   *     default this will be the version of Play Services that will be bundled
   */
  public AndroidForWorkAccountSupport(
      Context context, ComponentName admin, int minPlayServicesVersion) {
    this.context = context.getApplicationContext();
    this.admin = admin;
    this.minPlayServicesVersion = minPlayServicesVersion;
    Log.i(Constants.TAG, "Version: " + Constants.VERSION_CODE);
  }

  /**
   * Similar to {@link #ensureWorkingEnvironment(WorkingEnvironmentOptions,
   * WorkingEnvironmentCallback)}, but default options are used.
   */
  public void ensureWorkingEnvironment(final WorkingEnvironmentCallback callback) {
    ensureWorkingEnvironment(new WorkingEnvironmentOptions.Builder().build(), callback);
  }

  /**
   * Perform all necessary actions to add support for Android for Work Account provisioning. This
   * must be called by the Device Owner, Profile Owner or Device Admin app.
   *
   * <p>This includes:
   *
   * <ul>
   *   <li>Ensure Play Store and Play Services are sufficiently up to date, updating if possible
   *   <li>Remove any enroller accounts on the device if required
   *   <li>Enable the work account authenticator
   * </ul>
   *
   * <p>As part of this operation, {@link android.os.UserManager#ENSURE_VERIFY_APPS} is temporarily
   * set to avoid an additional dialogue from showing. After the operation has finished, this
   * setting is restored to the value it previously had. If this setting is changed externally
   * during this operation, it will be overwritten.
   *
   * <p>The result of this operation is reported through the {@code callback} on the UI thread.
   *
   * <p>When {@link WorkingEnvironmentCallback#onSuccess()} is called, the device is in a state
   * which supports Android for Work Accounts, after which {@link #addAndroidForWorkAccount(String,
   * WorkAccountAddedCallback)} can be called to add an account to the device.
   */
  public void ensureWorkingEnvironment(
      @NonNull final WorkingEnvironmentOptions option, final WorkingEnvironmentCallback callback) {
    ensureWorkingEnvironment(option, callback, new Handler(Looper.getMainLooper()));
  }

  /**
   * Perform all necessary actions to add support for Android for Work Account provisioning. Same as
   * {@link #ensureWorkingEnvironment} but dispatches callback events through the given handler.
   */
  private void ensureWorkingEnvironment(
      @NonNull final WorkingEnvironmentOptions options,
      final WorkingEnvironmentCallback callback,
      final Handler callbackHandler) {
    if (options == null) {
      throw new IllegalArgumentException("WorkingEnvironmentOptions should not be null");
    }

    HandlerThread handlerThread = new HandlerThread("ensureWorkingEnvironment-Handler");
    handlerThread.setUncaughtExceptionHandler(
        new UncaughtExceptionHandler() {
          @Override
          public void uncaughtException(Thread thread, final Throwable throwable) {
            Log.e(Constants.TAG, "Failed to ensure working environment. ", throwable);
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(WorkingEnvironmentCallback.Error.EXCEPTION, throwable);
                  }
                });
          }
        });
    final WorkingEnvironmentCallback workingEnvironmentCallback =
        new WorkingEnvironmentCallback() {
          @Override
          public void onSuccess() {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onSuccess();
                  }
                });
          }

          @Override
          public void onFailure(final Error error) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error);
                  }
                });
          }

          @Override
          public void onFailure(final Error error, final Throwable throwable) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error, throwable);
                  }
                });
          }

          @Override
          public void onProgressChange(final float progress) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onProgressChange(progress);
                  }
                });
          }
        };
    handlerThread.start();
    final Handler backgroundHandler = new Handler(handlerThread.getLooper());
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            new EnvironmentPreparer(context, admin, backgroundHandler, minPlayServicesVersion)
                .prepareEnvironment(workingEnvironmentCallback, options);
          }
        });
  }

  /**
   * Adds an Android for Work Account to the device using the specified token. Before this is
   * called, the work account authenticator should be enabled. This can be done by calling {@link
   * AndroidForWorkAccountSupport#ensureWorkingEnvironment(WorkingEnvironmentCallback)}.
   *
   * <p>The token is the authentication token generated by the EMM API
   * Users.GenerateAuthenticationToken call. This token should only ever be used once. If this call
   * fails or is retried for any reason, a new token should be used.
   *
   * <p>When the account is ready to use, the result is reported in {@link
   * WorkAccountAddedCallback#onAccountReady} on the UI thread. If there was an error while adding
   * the account, this will be reported in {@link
   * WorkAccountAddedCallback#onFailure(WorkAccountAddedCallback.Error)}.
   */
  public void addAndroidForWorkAccount(
      final String token, final WorkAccountAddedCallback callback) {
    ensurePlayServicesClientVersion();
    addAndroidForWorkAccount(token, callback, new Handler(Looper.getMainLooper()));
  }

  /**
   * Adds an Android for Work Account to the device using the specified token. Same as {@link
   * #addAndroidForWorkAccount} but dispatches callback events through the given handler.
   */
  private void addAndroidForWorkAccount(
      final String token, final WorkAccountAddedCallback callback, final Handler callbackHandler) {
    HandlerThread handlerThread = new HandlerThread("addAndroidForWorkAccount-Handler");
    handlerThread.setUncaughtExceptionHandler(
        new UncaughtExceptionHandler() {
          @Override
          public void uncaughtException(Thread thread, final Throwable throwable) {
            Log.e(Constants.TAG, "Failed to add account. ", throwable);
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(
                        WorkAccountAddedCallback.Error.EXCEPTION_ADDING_ACCOUNT, throwable);
                  }
                });
          }
        });
    final WorkAccountAddedCallback workAccountAddedCallback =
        new WorkAccountAddedCallback() {
          @Override
          public void onAccountReady(final Account account, final String deviceHint) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onAccountReady(account, deviceHint);
                  }
                });
          }

          @Override
          public void onFailure(final WorkAccountAddedCallback.Error error) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error);
                  }
                });
          }

          @Override
          public void onFailure(
              final WorkAccountAddedCallback.Error error, final Throwable throwable) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error, throwable);
                  }
                });
          }
        };
    handlerThread.start();
    final Handler backgroundHandler = new Handler(handlerThread.getLooper());
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            new WorkAccountProvisioner(context, admin)
                .addWorkAccount(token, minPlayServicesVersion, workAccountAddedCallback);
          }
        });
  }

  /**
   * Removes all Android for Work Accounts from the device. Before this is called, {@link
   * AndroidForWorkAccountSupport#ensureWorkingEnvironment(WorkingEnvironmentCallback)} should be
   * called.
   *
   * <p>The result of this operation is reported through the {@code callback} on the UI thread.
   */
  public void removeAllAndroidForWorkAccounts(final WorkAccountsRemovedCallback callback) {
    ensurePlayServicesClientVersion();
    removeAllAndroidForWorkAccounts(callback, new Handler(Looper.getMainLooper()));
  }

  /**
   * Same as {@link #removeAllAndroidForWorkAccounts} but dispatches callback events through the
   * given handler.
   */
  private void removeAllAndroidForWorkAccounts(
      final WorkAccountsRemovedCallback callback, final Handler callbackHandler) {
    HandlerThread handlerThread = new HandlerThread("removeAllAndroidForWorkAccounts-Handler");
    handlerThread.setUncaughtExceptionHandler(
        new UncaughtExceptionHandler() {
          @Override
          public void uncaughtException(Thread thread, final Throwable throwable) {
            Log.e(Constants.TAG, "Failed to remove accounts. ", throwable);
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(
                        WorkAccountsRemovedCallback.Error.EXCEPTION_REMOVING_ACCOUNT, throwable);
                  }
                });
          }
        });
    final WorkAccountsRemovedCallback workAccountsRemovedCallback =
        new WorkAccountsRemovedCallback() {
          @Override
          public void onSuccess() {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onSuccess();
                  }
                });
          }

          @Override
          public void onFailure(final WorkAccountsRemovedCallback.Error error) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error);
                  }
                });
          }

          @Override
          public void onFailure(
              final WorkAccountsRemovedCallback.Error error, final Throwable throwable) {
            callbackHandler.post(
                new Runnable() {
                  @Override
                  public void run() {
                    callback.onFailure(error, throwable);
                  }
                });
          }
        };
    handlerThread.start();
    final Handler backgroundHandler = new Handler(handlerThread.getLooper());
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            new WorkAccountRemover(context)
                .removeAllWorkAccounts(minPlayServicesVersion, workAccountsRemovedCallback);
          }
        });
  }

  private void ensurePlayServicesClientVersion() {
    if (GoogleApiAvailability.GOOGLE_PLAY_SERVICES_VERSION_CODE < PLAY_SERVICES_MIN_VERSION) {
      throw new RuntimeException(
          "Please update Google Play Services client library to "
              + PLAY_SERVICES_MIN_VERSION_READABLE);
    }
  }
}
