/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Handler;

/**
 * A one time only observer on a intent filter with a timeout period. {@link #onObserved} is called
 * when any of the intents in the intent filter is received.
 */
abstract class BroadcastObserver {

  private final Context context;
  private final Handler backgroundHandler;
  private final IntentFilter intentFilter;
  private final long timeoutMillis;

  private BroadcastReceiver broadcastReceiver;
  private boolean resultDelivered;

  /**
   * @param context the context associated with the caller
   * @param backgroundHandler the handler used to call {@link #onObserved} and {@link #onTimeout}
   * @param timeoutMillis time in millisecond before the observer stop observing the receiver
   */
  BroadcastObserver(
      Context context,
      final Handler backgroundHandler,
      IntentFilter intentFilter,
      long timeoutMillis) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
    this.intentFilter = intentFilter;
    this.timeoutMillis = timeoutMillis;
    this.broadcastReceiver =
        new BroadcastReceiver() {
          @Override
          public void onReceive(Context context, final Intent intent) {
            if (BroadcastObserver.this.intentFilter.hasAction(intent.getAction())) {
              // Run on background thread.
              backgroundHandler.post(
                  new Runnable() {
                    @Override
                    public void run() {
                      deliverResult(false /* isTimedOut */, intent);
                    }
                  });
            }
          }
        };
    this.resultDelivered = false;
  }

  /**
   * Delivers the first broadcast {@link Intent} matching the given {@link IntentFilter}. This
   * intent will be non-null.
   */
  abstract void onObserved(Intent observedIntent);

  /**
   * Called when the given timeout has passed.
   */
  abstract void onTimeout();

  /**
   * Start listening for intents.
   *
   * When the first intent matched by the provided {@link IntentFilter} is received,
   * {@link #onObserved} will be called.
   *
   * Exactly one of {@link #onObserved} and {@link #onTimeout} will be called.
   */
  void startObserving() {
    context.registerReceiver(broadcastReceiver, intentFilter);
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            deliverResult(true /* isTimedOut */, null);
          }
        },
        timeoutMillis);
  }

  private void stopObserving() {
    // Clear the receiver and timeout handler.
    if (broadcastReceiver != null) {
      context.unregisterReceiver(broadcastReceiver);
      broadcastReceiver = null;
    }
  }

  private synchronized void deliverResult(boolean isTimedOut, Intent intent) {
    if (!resultDelivered) {
      resultDelivered = true;
      stopObserving();
      if (isTimedOut) {
        onTimeout();
      } else {
        onObserved(intent);
      }
    }
  }
}
