/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.database.Cursor;
import android.net.Uri;
import android.os.Handler;
import android.text.TextUtils;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import java.util.concurrent.TimeUnit;

/**
 * The android checkin is an important dependency on many operations, such as adding an account to
 * the device.
 *
 * <p>This provides a helper method which is able to check whether the device has already checked
 * in, and wait until it does otherwise.
 */
class CheckinHelper {

  /** The location where the android id is stored in the local database. */
  private static final Uri GSERVICES_URI = Uri.parse("content://com.google.android.gsf.gservices");

  private static final String GSERVICES_KEY_ANDROID_ID = "android_id";

  private static final IntentFilter CHECKIN_COMPLETE_INTENT_FILTER =
      new IntentFilter(Constants.CHECKIN_COMPLETE_ACTION);
  @VisibleForTesting static final long CHECKIN_TIMEOUT_MILLIS = TimeUnit.MINUTES.toMillis(2);

  private final Context context;
  private final Handler backgroundHandler;

  CheckinHelper(Context context, Handler backgroundHandler) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
  }

  /**
   * Ensures that when {@link WorkingEnvironmentCallback#onSuccess()} is called, a device checkin
   * has happened. It is safe to call this method after a checkin has already happened, and it will
   * return immediately.
   *
   * <p>Google Services Framework stores the android id in a local database once the checkin has
   * succeeded, which is used to determine whether the checkin has happened.
   *
   * <p>{@link WorkingEnvironmentCallback#onFailure(WorkingEnvironmentCallback.Error)} is called if
   * this times out, or if there is an issue retrieving the android id from the device.
   */
  void ensureCheckinComplete(final WorkingEnvironmentCallback callback) {
    if (androidIdExists()) {
      Log.i(Constants.TAG, "Device has already done checkin.");
      callback.onSuccess();
      return;
    }
    Log.i(Constants.TAG, "Ensuring checkin is complete.");

    new BroadcastObserver(
        context, backgroundHandler, CHECKIN_COMPLETE_INTENT_FILTER, CHECKIN_TIMEOUT_MILLIS) {
      @Override
      void onObserved(Intent observedIntent) {
        if (androidIdExists()) {
          Log.i(Constants.TAG, "Checkin completed successfully.");
          callback.onSuccess();
        } else {
          Log.e(Constants.TAG, "Checkin complete but no android id found.");
          callback.onFailure(Error.CHECKIN_COMPLETE_BUT_NO_ANDROID_ID_FOUND);
        }
      }

      @Override
      void onTimeout() {
        if (androidIdExists()) {
          // This is logged as a warning as it shouldn't really happen, as it means there was a
          // checkin but we missed the broadcast.
          Log.w(Constants.TAG, "Checkin completed after timeout.");
          callback.onSuccess();
        } else {
          Log.e(Constants.TAG, "Timeout waiting for checkin.");
          callback.onFailure(Error.CHECKIN_TIMEOUT);
        }
      }
    }.startObserving();
  }

  /**
   * Retrieves the android id from the Google Services Framework. If the android id exists, checkin
   * has happened.
   *
   * <p>Returns true if checkin has happened.
   */
  private boolean androidIdExists() {
    String[] params = {GSERVICES_KEY_ANDROID_ID};
    Cursor cursor = null;
    try {
      cursor = context.getContentResolver().query(GSERVICES_URI, null, null, params, null);
      if (cursor == null) {
        return false;
      }
      if (!cursor.moveToFirst() || cursor.getColumnCount() < 2) {
        return false;
      }
      String androidId = cursor.getString(1);
      return !TextUtils.isEmpty(androidId);
    } catch (NumberFormatException e) {
      Log.e(Constants.TAG, "Exception parsing android id.", e);
      return false;
    } finally {
      if (cursor != null) {
        cursor.close();
      }
    }
  }
}
