package com.google.android.apps.work.dpcsupport;

import android.annotation.TargetApi;
import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.os.Build;
import android.os.Bundle;
import android.util.Log;

/**
 * Provides support for the device-wide unknown sources block.
 *
 * <p>If set within a work profile, this blocks unknown sources installations across all other users
 * on the device.
 *
 * <p>This is similar {@link android.os.UserManager#DISALLOW_INSTALL_UNKNOWN_SOURCES}, which also
 * blocks unknown sources, but on a per-user basis.
 */
@TargetApi(Build.VERSION_CODES.LOLLIPOP)
public class DeviceWideUnknownSourcesBlock {

  static final String DEVICE_WIDE_UNKNOWN_SOURCES_RESTRICTION =
      "verify_apps:device_wide_unknown_source_block";
  static final int MIN_SUPPORTED_PLAY_STORE_VERSION = 80812500;

  private final Context context;
  private final ComponentName adminComponent;

  public DeviceWideUnknownSourcesBlock(Context context, ComponentName adminComponent) {
    if (context == null) {
      throw new NullPointerException("Context cannot be null");
    }
    if (adminComponent == null) {
      throw new NullPointerException("Admin component cannot be null");
    }
    this.context = context;
    this.adminComponent = adminComponent;
  }

  /** Enables the device wide unknown sources block by setting the corresponding restriction. */
  public void enable() {
    DevicePolicyManager devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    Bundle originalBundle =
        devicePolicyManager.getApplicationRestrictions(
            adminComponent, Constants.PLAY_STORE_PACKAGE_NAME);
    Bundle newBundle = new Bundle(originalBundle);
    newBundle.putBoolean(DEVICE_WIDE_UNKNOWN_SOURCES_RESTRICTION, true);
    devicePolicyManager.setApplicationRestrictions(
        adminComponent, Constants.PLAY_STORE_PACKAGE_NAME, newBundle);
  }

  /** Disables the device wide unknown sources block by setting the corresponding restriction. */
  public void disable() {
    DevicePolicyManager devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    Bundle originalBundle =
        devicePolicyManager.getApplicationRestrictions(
            adminComponent, Constants.PLAY_STORE_PACKAGE_NAME);
    Bundle newBundle = new Bundle(originalBundle);
    newBundle.putBoolean(DEVICE_WIDE_UNKNOWN_SOURCES_RESTRICTION, false);
    devicePolicyManager.setApplicationRestrictions(
        adminComponent, Constants.PLAY_STORE_PACKAGE_NAME, newBundle);
  }

  /**
   * Returns whether the unknown sources block is currently enforced. For this to be enforced, the
   * restriction must be set with {@link #enable()}, and the Play Store must be sufficiently up to
   * date.
   */
  public boolean isActive() {
    DevicePolicyManager devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    Bundle activeRestrictions =
        devicePolicyManager.getApplicationRestrictions(
            adminComponent, Constants.PLAY_STORE_PACKAGE_NAME);
    boolean restrictionActive =
        activeRestrictions.getBoolean(DEVICE_WIDE_UNKNOWN_SOURCES_RESTRICTION);
    return restrictionActive && isPlayStoreAtSupportedVersion();
  }

  private boolean isPlayStoreAtSupportedVersion() {
    PackageManager packageManager = context.getPackageManager();
    int installedPlayStoreVersion;
    try {
      installedPlayStoreVersion =
          packageManager.getPackageInfo(Constants.PLAY_STORE_PACKAGE_NAME, 0).versionCode;
    } catch (NameNotFoundException e) {
      Log.e(Constants.TAG, "Play store not found.", e);
      return false;
    }
    return installedPlayStoreVersion >= MIN_SUPPORTED_PLAY_STORE_VERSION;
  }
}
