/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.accounts.Account;
import android.accounts.AccountManager;
import android.app.admin.DevicePolicyManager;
import android.content.Context;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.SystemClock;
import android.os.UserManager;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.gms.auth.GoogleAuthException;
import com.google.android.gms.auth.GoogleAuthUtil;
import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * As part of the Android for Work Device Owner provisioning process, a temporary "enroller" account
 * is added to the device. The sole purpose of this account is to download the correct DPC, after
 * which it needs to be removed from the device.
 */
class EnrollerAccountRemover {

  private static final long REMOVE_ENROLLER_ACCOUNT_TIMEOUT_SECONDS = 30;

  /**
   * The interval between remove enroller account attempts.
   * Note that this timeout value must be greater than
   * GmsClientSupervisorImpl#DEFAULT_UNBIND_DELAY_MILLIS.
   * See go/failed_remove_enroller_accounts for details.
   */
  private static final long REMOVE_ENROLLER_ACCOUNT_RETRY_WAIT_INTERVAL_MS =
      (int) TimeUnit.MILLISECONDS.convert(10, TimeUnit.SECONDS);

  /** The limit of removing enroller account retry */
  @VisibleForTesting static final int REMOVE_ENROLLER_ACCOUNT_RETRY_LIMIT = 50;

  private final Context context;
  private final DevicePolicyManagerHelper devicePolicyManagerHelper;
  private final DevicePolicyManager devicePolicyManager;
  private final UserManager userManager;

  EnrollerAccountRemover(Context context) {
    this.context = context;
    devicePolicyManagerHelper = new DevicePolicyManagerHelper(context);
    this.devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    this.userManager = (UserManager) context.getSystemService(Context.USER_SERVICE);
  }

  /**
   * Removes all enroller accounts on a device which were added as part of provisioning. This should
   * only be called at the beginning of Android for Work Device Owner setup.
   *
   * <p>Must only be called by device owner.
   *
   * <p>Play services must be at least v3 to support this.
   *
   * @return true if enroller accounts (if any) were successfully removed.
   */
  @WorkerThread
  boolean removeEnrollerAccounts() {
    AccountManager accountManager = AccountManager.get(context);
    Account[] accountsOnDevice =
        accountManager.getAccountsByType(GoogleAuthUtil.GOOGLE_ACCOUNT_TYPE);
    for (Account account : accountsOnDevice) {
      if (account.name.contains("@")) {
        // The enroller account added is not a real account - its name is a non-email address string
        // We expect this method to be called as the first thing in a AfW DO setup. However, if it
        // is mistakenly called at a later point, we add this sanity check in to ensure we do not
        // remove real accounts.
        Log.w(Constants.TAG, "Skipping removal of non-enroller account: " + account.name);
        continue;
      }
      Log.i(Constants.TAG, "Removing enroller account: " + account.name);
      boolean success = false;
      for (int count = 0; count < REMOVE_ENROLLER_ACCOUNT_RETRY_LIMIT && !success; count++) {
        success = removeAccount(account);
        if (count < REMOVE_ENROLLER_ACCOUNT_RETRY_LIMIT && !success) {
          SystemClock.sleep(REMOVE_ENROLLER_ACCOUNT_RETRY_WAIT_INTERVAL_MS);
        }
      }
      if (!success) {
        return false;
      }
    }
    return true; // There are now no enroller accounts left on device
  }

  private boolean removeAccount(Account account) {
    if (!devicePolicyManagerHelper.isDeviceOwner()) {
      Log.e(Constants.TAG, "Cannot remove enroller accounts when not device owner.");
      return false;
    }
    if (Arrays.asList(devicePolicyManager.getAccountTypesWithManagementDisabled())
        .contains(GoogleAuthUtil.GOOGLE_ACCOUNT_TYPE)) {
      Log.e(
          Constants.TAG,
          "Cannot remove enroller accounts if Google account management is disabled");
      return false;
    }
    if (userManager.hasUserRestriction(UserManager.DISALLOW_MODIFY_ACCOUNTS)) {
      Log.e(Constants.TAG, "Cannot remove enroller accounts disallowed modifying accounts");
      return false;
    }
    AsyncTask<Account, Void, Boolean> removeAccountTask =
        new AsyncTask<Account, Void, Boolean>() {
          @Override
          protected Boolean doInBackground(Account... params) {
            Account account = params[0];
            try {
              // This call allows the DO to remove accounts. Play Services must have version
              // 8400000 or above to use this API, which has already been checked by
              // {@code EnvironmentPreparer}.
              Bundle result = GoogleAuthUtil.removeAccount(context, account);
              return result.getBoolean(AccountManager.KEY_BOOLEAN_RESULT);
            } catch (GoogleAuthException | IOException e) {
              Log.e(Constants.TAG, "Failed to remove enroller account: " + account.name, e);
              return false;
            }
          }
        }.execute(account);
    try {
      return removeAccountTask.get(REMOVE_ENROLLER_ACCOUNT_TIMEOUT_SECONDS, TimeUnit.SECONDS);
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      Log.e(Constants.TAG, "Failed to remove enroller account - thread interrupted", e);
    } catch (ExecutionException e) {
      Log.e(Constants.TAG, "Failed to remove enroller account - failed", e);
    } catch (TimeoutException e) {
      Log.e(Constants.TAG, "Failed to remove enroller account - timed out", e);
      removeAccountTask.cancel(true /* mayInterruptIfRunning */);
    }
    return false;
  }
}
