/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.content.Context;
import android.os.Handler;
import android.os.SystemClock;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import java.util.concurrent.TimeUnit;

/** Updates package using {@link PackageUpdater} and retries on failures. */
class PackageUpdateScheduler {
  /**
   * The download size is for Play Services is ~50mb, we'll use that as timeout for any packages.
   */
  @VisibleForTesting static final long PACKAGE_UPDATE_TIMEOUT_MS = TimeUnit.HOURS.toMillis(20);

  /** Delay after which the update is not retried */
  @VisibleForTesting
  static final long PACKAGE_UPDATE_START_TIMEOUT_MS = TimeUnit.MINUTES.toMillis(2);

  @VisibleForTesting static final long PACKAGE_UPDATE_RETRY_WAIT_MS = TimeUnit.SECONDS.toMillis(10);

  private final Context context;
  private final Handler backgroundHandler;

  private WorkingEnvironmentCallback callback;
  private String packageName;

  private long maxPackageUpdateStartTimeMs;
  private boolean updateComplete = false;

  PackageUpdateScheduler(Context context, Handler backgroundHandler) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
  }

  /** Updates package, and retries on failures. */
  @WorkerThread
  void updatePackage(
      WorkingEnvironmentCallback callback, String packageName, WorkingEnvironmentOptions options) {
    this.callback = callback;
    this.packageName = packageName;

    activateTimeout();

    tryUpdatePackage(options);
  }

  @WorkerThread
  private void tryUpdatePackage(final WorkingEnvironmentOptions options) {
    final WorkingEnvironmentCallback updateCompleteCallback =
        new WorkingEnvironmentCallback() {
          @Override
          public void onSuccess() {
            success();
          }

          @Override
          public void onFailure(Error statusCode) {
            retry(statusCode, options);
          }

          @Override
          public void onProgressChange(float progress) {
            callback.onProgressChange(progress);
          }
        };

    new PackageUpdater(context, backgroundHandler)
        .updatePackage(updateCompleteCallback, packageName, options);
  }

  private void activateTimeout() {
    maxPackageUpdateStartTimeMs = SystemClock.uptimeMillis() + PACKAGE_UPDATE_START_TIMEOUT_MS;

    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            failure(Error.PLAY_SERVICES_UPDATE_TIMEOUT);
          }
        },
        PACKAGE_UPDATE_TIMEOUT_MS);
  }

  private void success() {
    if (updateComplete) {
      return;
    }
    Log.i(Constants.TAG, "PackageUpdateScheduler Package successfully updated: " + packageName);
    updateComplete = true;
    callback.onSuccess();
  }

  private void retry(Error statusCode, final WorkingEnvironmentOptions options) {
    if (updateComplete) {
      return;
    }
    if (SystemClock.uptimeMillis() > maxPackageUpdateStartTimeMs) {
      failure(statusCode);
      return;
    }

    Log.i(Constants.TAG, "Retrying Package update: " + packageName);
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            tryUpdatePackage(options);
          }
        },
        PACKAGE_UPDATE_RETRY_WAIT_MS);
  }

  private void failure(Error reason) {
    if (updateComplete) {
      return;
    }
    updateComplete = true;
    callback.onFailure(reason);
  }
}
