/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.pm.PackageInstaller;
import android.content.pm.PackageInstaller.SessionCallback;
import android.content.pm.PackageInstaller.SessionInfo;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import com.google.android.finsky.services.IPackageUpdateService;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

/** Updates package. */
class PackageUpdater {

  @VisibleForTesting
  static final String PACKAGE_UPDATE_SERVICE_INTENT =
      "com.google.android.finsky.BIND_PACKAGE_UPDATE_SERVICE";

  @VisibleForTesting
  static final long PACKAGE_UPDATE_SESSION_TIMEOUT_MS =
      TimeUnit.MILLISECONDS.convert(30, TimeUnit.SECONDS);

  private static final long PACKAGE_UPDATE_SERVICE_CONNECTION_TIMEOUT_MS =
      TimeUnit.MILLISECONDS.convert(10, TimeUnit.SECONDS);

  // Result from remote {@link IPackageUpdateService}
  private static final String KEY_SUCCESS = "success";
  private static final String KEY_UNAUTHENTICATED = "unauthenticated";

  private final Context context;
  private final Handler backgroundHandler;
  private final PackageInstaller packageInstaller;

  private final List<Integer> packageUpdateSessions = new ArrayList<>();
  private final SessionCallback sessionCallback = new InstallSessionCallback();

  private WorkingEnvironmentCallback callback;
  private String packageName;

  private boolean serviceConnected = false;
  private boolean updateComplete = false;

  PackageUpdater(Context context, Handler backgroundHandler) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
    this.packageInstaller = context.getPackageManager().getPackageInstaller();
  }

  /** Updates package to a version supporting Android for Work Accounts. */
  @WorkerThread
  void updatePackage(
      WorkingEnvironmentCallback callback, String packageName, WorkingEnvironmentOptions options) {
    this.callback = callback;
    this.packageName = packageName;

    // Register the callback on the main thread.
    packageInstaller.registerSessionCallback(sessionCallback, new Handler(Looper.getMainLooper()));

    // There might already be an ongoing packageInstaller session.
    for (SessionInfo sessionInfo : packageInstaller.getAllSessions()) {
      if (packageName.equals(sessionInfo.getAppPackageName())) {
        packageUpdateSessions.add(sessionInfo.getSessionId());
      }
    }
    // If there is one, there is no need to call the API.
    if (!packageUpdateSessions.isEmpty()) {
      Log.i(Constants.TAG, "packageUpdateSessions is not empty");
      return;
    }

    Intent intent = new Intent(PACKAGE_UPDATE_SERVICE_INTENT);
    intent.setPackage(Constants.PLAY_STORE_PACKAGE_NAME);

    // Bind the service.
    if (!context.bindService(
        intent, new PackageUpdateServiceConnection(options), Context.BIND_AUTO_CREATE)) {
      Log.e(Constants.TAG, "Unable to connect to service");
      failure(Error.PLAY_SERVICES_UPDATE_CONNECTION_FAILED);
    }

    activateServiceConnectionTimeout();
  }

  @WorkerThread
  private void callPackageUpdateService(
      ServiceConnection connection, IBinder service, WorkingEnvironmentOptions options) {
    IPackageUpdateService packageUpdateService = IPackageUpdateService.Stub.asInterface(service);
    serviceConnected = true;

    try {
      Bundle extrasBundle = new Bundle();
      // Indicates that no account should be used
      extrasBundle.putBoolean(KEY_UNAUTHENTICATED, !options.useAuthenticatedPackageUpdate);
      Bundle result = packageUpdateService.updatePackage(packageName, extrasBundle);
      context.unbindService(connection);

      if (!result.getBoolean(KEY_SUCCESS, false)) {
        Log.e(Constants.TAG, "Package update failed to start: " + packageName);
        failure(Error.PLAY_SERVICES_UPDATE_FAILED_TO_START);
      }
    } catch (Exception e) {
      Log.e(Constants.TAG, "Failure in package update service.", e);
      failure(Error.PLAY_SERVICES_UPDATE_REMOTE_FAILURE);
    }

    activateSessionCreatedTimeout();
  }

  private void activateServiceConnectionTimeout() {
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            if (!serviceConnected) {
              Log.e(Constants.TAG, "Timed out while connecting to service.");
              failure(Error.PLAY_SERVICES_UPDATE_CONNECTION_FAILED);
            }
          }
        },
        PACKAGE_UPDATE_SERVICE_CONNECTION_TIMEOUT_MS);
  }

  private void activateSessionCreatedTimeout() {
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            if (packageUpdateSessions.isEmpty()) {
              failure(Error.PLAY_SERVICES_SESSION_NOT_FOUND);
            }
          }
        },
        PACKAGE_UPDATE_SESSION_TIMEOUT_MS);
  }

  private void success() {
    if (updateComplete) {
      return;
    }
    Log.i(Constants.TAG, "Package successfully updated: " + packageName);
    updateComplete = true;
    packageInstaller.unregisterSessionCallback(sessionCallback);
    callback.onSuccess();
  }

  private void failure(Error reason) {
    if (updateComplete) {
      return;
    }
    updateComplete = true;
    packageInstaller.unregisterSessionCallback(sessionCallback);
    callback.onFailure(reason);
  }

  private class PackageUpdateServiceConnection implements ServiceConnection {

    private final WorkingEnvironmentOptions options;

    private PackageUpdateServiceConnection(WorkingEnvironmentOptions options) {
      this.options = options;
    }

    @Override
    public void onServiceConnected(ComponentName className, final IBinder service) {
      final ServiceConnection connection = this;
      // Run on background thread
      backgroundHandler.post(
          new Runnable() {
            @Override
            public void run() {
              callPackageUpdateService(connection, service, options);
            }
          });
    }

    @Override
    public void onServiceDisconnected(ComponentName className) {
      // The service is automatically disconnected when we unbind it after a successful call.
      Log.e(Constants.TAG, "Service has been disconnected");
    }
  }

  private class InstallSessionCallback extends SessionCallback {
    @Override
    public void onCreated(int sessionId) {
      SessionInfo sessionInfo = packageInstaller.getSessionInfo(sessionId);
      if (sessionInfo == null) {
        Log.e(Constants.TAG, "sessionInfo is null");
        return;
      }
      if (packageName.equals(sessionInfo.getAppPackageName())) {
        packageUpdateSessions.add(sessionId);
      }
    }

    @Override
    public void onBadgingChanged(int sessionId) {}

    @Override
    public void onActiveChanged(int sessionId, boolean active) {}

    @Override
    public void onProgressChanged(int sessionId, float progress) {
      if (packageUpdateSessions.contains(sessionId)) {
        Log.i(Constants.TAG, "Package update: " + packageName + " progress: " + progress);
        callback.onProgressChange(progress);
      }
    }

    @Override
    public void onFinished(final int sessionId, final boolean success) {
      // Run on background thread.
      backgroundHandler.post(
          new Runnable() {
            @Override
            public void run() {
              if (packageUpdateSessions.contains(sessionId)) {
                packageUpdateSessions.remove((Integer) sessionId);
                if (success) {
                  success();
                } else {
                  failure(Error.PLAY_SERVICES_PACKAGE_UPDATE_FAILED);
                }
              }
            }
          });
    }
  }
}
