/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.app.DownloadManager;
import android.app.DownloadManager.Request;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.content.pm.Signature;
import android.database.Cursor;
import android.net.Uri;
import android.os.Handler;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Downloads the correct (release or debug) play store apk from a public URL.
 */
class PlayStoreDownloader extends BroadcastReceiver {

  /** APKs are downloaded via the same mechanism as DPCs: */
  static final Uri PLAY_STORE_PROD_URI =
      Uri.parse("https://play.google.com/managed/downloadManagingApp?identifier=o1u2mkmd6taq");

  static final Uri PLAY_STORE_DEBUG_URI =
      Uri.parse("https://play.google.com/managed/downloadManagingApp?identifier=ljhflno0k3nm");

  private static final IntentFilter DOWNLOAD_COMPLETE_INTENT_FILTER =
      new IntentFilter(DownloadManager.ACTION_DOWNLOAD_COMPLETE);

  /** The download size is ~20mb. */
  @VisibleForTesting
  static final long PLAY_STORE_DOWNLOAD_TIMEOUT_MS =
      TimeUnit.MILLISECONDS.convert(10, TimeUnit.HOURS);

  @VisibleForTesting
  static final AtomicInteger DOWNLOAD_SUFFIX = new AtomicInteger(0);

  @VisibleForTesting
  static final String PLAY_STORE_LOCAL_DOWNLOAD_PATH =
      "/dpcsupport_download_cache/play-store-%s.apk";

  interface DownloadCompleteListener {
    void onComplete(InputStream inputStream);

    void onFailure(Error reason);

    void onProgressChange(float progress);
  }

  private final Context context;
  private final Handler backgroundHandler;
  private final PackageManager packageManager;
  private final DownloadManager downloadManager;
  private final String filePath;

  private DownloadCompleteListener listener;
  private Long downloadId;
  private boolean downloadFinished;
  private File downloadedFile;

  @VisibleForTesting
  PlayStoreDownloader(Context context, Handler backgroundHandler) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
    this.downloadManager = (DownloadManager) context.getSystemService(Context.DOWNLOAD_SERVICE);
    this.packageManager = context.getPackageManager();
    this.filePath =
        String.format(PLAY_STORE_LOCAL_DOWNLOAD_PATH, DOWNLOAD_SUFFIX.getAndIncrement());
  }

  /** Starts the download, reporting the result to the provided {@link DownloadCompleteListener}. */
  @WorkerThread
  void downloadPlayStore(final DownloadCompleteListener listener) {
    this.listener = listener;

    context.registerReceiver(this, DOWNLOAD_COMPLETE_INTENT_FILTER);
    Uri playStoreDownloadUri = getPlayStoreDownloadUri();
    if (playStoreDownloadUri == null) {
      return;
    }
    activateTimeout();
    String path = context.getApplicationContext().getExternalFilesDir(null) + filePath;
    downloadedFile = new File(path);
    downloadedFile.getParentFile().mkdirs(); // If the folder doesn't exists it is created

    DownloadManager.Request request =
        new DownloadManager.Request(playStoreDownloadUri)
            .setDestinationUri(Uri.fromFile(downloadedFile))
            .setVisibleInDownloadsUi(false)
            .setNotificationVisibility(Request.VISIBILITY_HIDDEN);
    downloadId = downloadManager.enqueue(request);
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            if (!downloadFinished) {
              float progress = getDownloadProgress();
              listener.onProgressChange(progress);
              backgroundHandler.postDelayed(this, 1000);
            }
          }
        });
  }

  /**
   * Provides the correct URI to download the Play Store APK from. This is determined by checking
   * the signature of the existing installed APK.
   */
  private Uri getPlayStoreDownloadUri() {
    Signature[] signatures;
    try {
      PackageInfo packageInfo =
          packageManager.getPackageInfo(
              Constants.PLAY_STORE_PACKAGE_NAME,
              PackageManager.GET_SIGNATURES);
      signatures = packageInfo.signatures;
    } catch (NameNotFoundException e) {
      Log.e(Constants.TAG, "Play Store was not installed", e);
      fail(Error.PLAY_STORE_NOT_FOUND);
      return null;
    }
    for (Signature signature : signatures) {
      if (signature.equals(Constants.PLAY_STORE_RELEASE_KEY)) {
        return PLAY_STORE_PROD_URI;
      } else if (signature.equals(Constants.PLAY_STORE_DEBUG_KEY)) {
        return PLAY_STORE_DEBUG_URI;
      }
    }
    Log.e(Constants.TAG, "No matching Play Store signature found.");
    fail(Error.PLAY_STORE_SIGNATURE_MISMATCH);
    return null;
  }

  @WorkerThread
  private void processCompletedDownload(Long completedDownloadId) {
    if (!completedDownloadId.equals(downloadId)) {
      return;
    }

    Cursor cursor = null;
    int status;
    int reason;
    try {
      cursor = downloadManager.query(new DownloadManager.Query().setFilterById(downloadId));
      cursor.moveToFirst();
      status = cursor.getInt(cursor.getColumnIndex(DownloadManager.COLUMN_STATUS));
      reason = cursor.getInt(cursor.getColumnIndex(DownloadManager.COLUMN_REASON));
    } finally {
      if (cursor != null) {
        cursor.close();
      }
    }

    if (status != DownloadManager.STATUS_SUCCESSFUL) {
      Log.e(Constants.TAG, "Download failed with reason " + reason + ".");
      fail(Error.PLAY_STORE_DOWNLOAD_FAILED);
      return;
    }

    FileInputStream inputStream;
    try {
      inputStream = new FileInputStream(downloadedFile);
      Log.i(Constants.TAG, "Completed Play Store download.");
    } catch (IOException e) {
      Log.e(Constants.TAG, "Failed to open play store apk", e);
      fail(Error.PLAY_STORE_DOWNLOAD_FAILED);
      return;
    }
    success(inputStream);
  }

  private void activateTimeout() {
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            fail(Error.PLAY_STORE_DOWNLOAD_TIMEOUT);
          }
        },
        PLAY_STORE_DOWNLOAD_TIMEOUT_MS);
  }

  private void cleanup() {
    Log.i(Constants.TAG, "Cleanup Play Store download");
    context.unregisterReceiver(this);
    if (downloadId != null) {
      downloadManager.remove(downloadId);
      downloadId = null;
    }
  }

  private void success(InputStream inputStream) {
    if (downloadFinished) {
      return;
    }
    downloadFinished = true;
    listener.onComplete(inputStream);
    cleanup(); // onComplete uses the inputStream synchronously
  }

  private void fail(Error errorCode) {
    if (downloadFinished) {
      return;
    }
    downloadFinished = true;
    listener.onFailure(errorCode);
    cleanup();
  }

  @Override
  public void onReceive(Context context, final Intent intent) {
    // Run on background thread.
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            processCompletedDownload(intent.getLongExtra(DownloadManager.EXTRA_DOWNLOAD_ID, -1));
          }
        });
  }

  private float getDownloadProgress() {
    DownloadManager.Query query = new DownloadManager.Query();
    query.setFilterById(downloadId);
    Cursor cursor = null;
    float progress = 0f;
    try {
      cursor = downloadManager.query(query);
      cursor.moveToFirst();
      int bytesDownloaded =
          cursor.getInt(cursor.getColumnIndex(DownloadManager.COLUMN_BYTES_DOWNLOADED_SO_FAR));
      int totalBytes =
          cursor.getInt(cursor.getColumnIndex(DownloadManager.COLUMN_TOTAL_SIZE_BYTES));
      if (totalBytes > 0) {
        progress = (float) bytesDownloaded / (float) totalBytes;
      }
    } finally {
      if (cursor != null) {
        cursor.close();
      }
    }
    return progress >= 0 ? progress : 0.0f;
  }

}
