package com.google.android.apps.work.dpcsupport;

import android.annotation.TargetApi;
import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.content.pm.PermissionInfo;
import android.content.pm.ServiceInfo;
import android.os.Build.VERSION;
import android.os.Build.VERSION_CODES;
import android.util.Log;
import java.util.Arrays;
import java.util.List;

/**
 * Helper class for checking preconditions during provisioning.
 */
class PreconditionsHelper {

  private final ComponentName admin;
  private final Context context;
  private final DevicePolicyManagerHelper devicePolicyManagerHelper;

  PreconditionsHelper(
      Context context, ComponentName admin, DevicePolicyManagerHelper devicePolicyManagerHelper) {
    this.admin = admin;
    this.context = context;
    this.devicePolicyManagerHelper = devicePolicyManagerHelper;
  }

  /**
   * Checks that the given permissions are present, and grants any dangerous permissions required.
   */
  boolean ensureHasPermissions(String[] requiredPermissions) {
    PackageInfo packageInfo = getPackageInfo(PackageManager.GET_PERMISSIONS);
    if (packageInfo == null) {
      return false;
    }

    List<String> manifestPermissions = Arrays.asList(packageInfo.requestedPermissions);
    for (String expectedPermission : requiredPermissions) {
      if (!manifestPermissions.contains(expectedPermission)) {
        Log.e(Constants.TAG, "Missing required permission from manifest: " + expectedPermission);
        return false;
      }
      maybeGrantDangerousPermission(expectedPermission);
    }
    return true;
  }

  /**
   * Checks that the given service is present and exported, and enables it.
   */
  boolean ensureHasService(String serviceName) {
    PackageInfo packageInfo = getPackageInfo(PackageManager.GET_SERVICES);
    if (packageInfo == null || packageInfo.services == null) {
      return false;
    }

    for (ServiceInfo declaredService : packageInfo.services) {
      if (declaredService.name.equals(serviceName)) {
        if (!declaredService.exported) {
          Log.e(Constants.TAG, String.format("Service %s not exported.", declaredService.name));
          return false;
        }

        ensureServiceEnabled(declaredService);
        return true;
      }
    }

    return false;
  }

  /**
   * Attempts to grant a permission automatically if it is considered dangerous - this only happens
   * for PO/DO devices.
   */
  private void maybeGrantDangerousPermission(String permission) {
    if (VERSION.SDK_INT >= VERSION_CODES.M
        && devicePolicyManagerHelper.isDeviceOrProfileOwner()
        && isPermissionDangerous(permission)
        && !hasPermissionGranted(admin, context.getPackageName(), permission)) {
      DevicePolicyManager devicePolicyManager =
          (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
      devicePolicyManager.setPermissionGrantState(
          admin,
          context.getPackageName(),
          permission,
          DevicePolicyManager.PERMISSION_GRANT_STATE_GRANTED);
    }
  }

  @TargetApi(VERSION_CODES.M)
  private boolean hasPermissionGranted(
      ComponentName componentName, String packageName, String permission) {
    DevicePolicyManager devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    return devicePolicyManager.getPermissionGrantState(componentName, packageName, permission)
        == DevicePolicyManager.PERMISSION_GRANT_STATE_GRANTED;
  }

  private boolean isPermissionDangerous(String permission) {
    PermissionInfo permissionInfo;
    try {
      permissionInfo = context.getPackageManager().getPermissionInfo(permission, 0);
    } catch (NameNotFoundException e) {
      Log.e(Constants.TAG, "Failed to look up permission.", e);
      return false;
    }
    return permissionInfo != null
        && (permissionInfo.protectionLevel & PermissionInfo.PROTECTION_MASK_BASE)
            == PermissionInfo.PROTECTION_DANGEROUS;
  }

  private PackageInfo getPackageInfo(int flags) {
    try {
      return context
          .getPackageManager()
          .getPackageInfo(context.getPackageName(), flags);
    } catch (NameNotFoundException e) {
      Log.e(Constants.TAG, "Could not find own package.", e);
      return null;
    }
  }

  private void ensureServiceEnabled(ServiceInfo declaredService) {
    if (declaredService.enabled) {
      return;
    }

    Log.i(Constants.TAG, String.format("Enabling service %s.", declaredService.name));
    context
        .getPackageManager()
        .setComponentEnabledSetting(
            new ComponentName(context, declaredService.name),
            PackageManager.COMPONENT_ENABLED_STATE_ENABLED,
            PackageManager.DONT_KILL_APP);
  }
}
