/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.accounts.Account;
import android.accounts.AccountManager;
import android.content.Context;
import android.os.SystemClock;
import android.util.Log;
import com.google.android.gms.auth.GoogleAuthUtil;
import com.google.android.gms.auth.account.WorkAccount;
import com.google.android.gms.auth.account.WorkAccountClient;
import com.google.android.gms.tasks.Task;
import com.google.android.gms.tasks.Tasks;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Interacts with the Play Services work account API.
 */
class WorkAccountApiHelper {

  private static final int ENABLE_WORK_AUTHENTICATOR_TIMEOUT_SECONDS = 5;

  private static final int REMOVE_WORK_AUCCOUNT_TIMEOUT_SECONDS = 5;

  private static final int ADD_WORK_ACCOUNT_TIMEOUT_SECONDS = 20;

  /** The interval between attempts */
  private static final long RETRY_WAIT_INTERVAL_MS = TimeUnit.SECONDS.toMillis(1);

  /** The limit of retries */
  private static final int RETRY_LIMIT = 5;

  private final WorkAccountClient client;

  private AccountManager accountManager;

  WorkAccountApiHelper(Context context) {
    client = WorkAccount.getClient(context);
    accountManager = AccountManager.get(context);
  }

  /**
   * Enables the work account authenticator.
   */
  void enableWorkAuthenticator() {
    for (int count = 0; count < RETRY_LIMIT; count++) {
      Log.i(Constants.TAG, String.format("Enabling work authenticator, attempt %d", count));

      Task<Void> task = client.setWorkAuthenticatorEnabled(true);
      awaitTask(task, ENABLE_WORK_AUTHENTICATOR_TIMEOUT_SECONDS);
      if (task.isSuccessful()) {
        return;
      }

      SystemClock.sleep(RETRY_WAIT_INTERVAL_MS);
    }
  }

  /** Adds an Android for Work account. */
  Account addWorkAccount(String token) throws Exception {
    Account workAccount = getFirstWorkAccount();
    if (workAccount != null) {
      throw new IllegalStateException("Trying to add another Managed Google Play Account");
    }

    for (int count = 0; count < RETRY_LIMIT; count++) {
      Log.i(Constants.TAG, String.format("Adding work account, attempt %d", count));

      Task<Account> task = client.addWorkAccount(token);
      awaitTask(task, ADD_WORK_ACCOUNT_TIMEOUT_SECONDS);
      if (task.isSuccessful()) {
        return task.getResult();
      }
      // Maybe the account is already added, but GmsCore just failed to deliver the result
      // as being killed.
      workAccount = getFirstWorkAccount();
      if (workAccount != null) {
        Log.i(Constants.TAG, "Work account is found: " + workAccount);
        return workAccount;
      }
      Exception exception = task.getException();
      if (count == RETRY_LIMIT - 1 && exception != null) {
        throw exception;
      } else {
        SystemClock.sleep(RETRY_WAIT_INTERVAL_MS);
      }
    }
    return null;
  }

  /**
   * Removes an Android for Work account.
   */
  boolean removeWorkAccount(Account account) {
    for (int count = 0; count < RETRY_LIMIT; count++) {
      Log.i(
          Constants.TAG,
          String.format("Removing work account %s, attempt %d", account.name, count));

      Task<Void> task = client.removeWorkAccount(account);
      awaitTask(task, REMOVE_WORK_AUCCOUNT_TIMEOUT_SECONDS);
      if (task.isSuccessful()) {
        return true;
      }

      SystemClock.sleep(RETRY_WAIT_INTERVAL_MS);
    }
    return false;
  }

  private void awaitTask(Task<?> task, int timeoutSeconds) {
    try {
      Tasks.await(task, timeoutSeconds, TimeUnit.SECONDS);
    } catch (ExecutionException | InterruptedException | TimeoutException e) {
      if (e instanceof InterruptedException) {
        Thread.currentThread().interrupt();
      }
      Log.e(Constants.TAG, "Attempt failed with ", e);
    }
  }

  private Account getFirstWorkAccount() {
    Account[] workAccounts = accountManager.getAccountsByType(GoogleAuthUtil.WORK_ACCOUNT_TYPE);
    if (workAccounts == null || workAccounts.length == 0) {
      return null;
    }
    return workAccounts[0];
  }
}
