/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.accounts.AccountManager;
import android.accounts.AuthenticatorDescription;
import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.os.Bundle;
import android.os.SystemClock;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import com.google.android.gms.common.GoogleApiAvailability;
import java.util.concurrent.TimeUnit;

/** Enables the Android for Work Account authenticator. */
class WorkAccountAuthenticatorEnabler {

  /** The interval between polling whether the authenticator was enabled */
  @VisibleForTesting static final int ENABLE_WORK_AUTHENTICATOR_WAIT_INTERVAL_MS =
      (int) TimeUnit.MILLISECONDS.convert(1, TimeUnit.SECONDS);

  /** The maximum time we wait for the authenticator to be enabled. */
  @VisibleForTesting static final int ENABLE_WORK_AUTHENTICATOR_TIMEOUT_MS =
      (int) TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES);

  @VisibleForTesting
  static final String ENABLE_WORK_ACCOUNTS_RESTRICTION = "enableWorkAccountAdmin";

  private final Context context;
  private final ComponentName admin;
  private final DevicePolicyManagerHelper devicePolicyManagerHelper;
  private final WorkAccountApiHelper workAccountApiHelper;

  WorkAccountAuthenticatorEnabler(
      Context context,
      ComponentName admin,
      DevicePolicyManagerHelper devicePolicyManagerHelper,
      WorkAccountApiHelper workAccountApiHelper) {
    this.context = context;
    this.admin = admin;
    this.devicePolicyManagerHelper = devicePolicyManagerHelper;
    this.workAccountApiHelper = workAccountApiHelper;
  }

  /**
   * Enables the work account authenticator. Returns null if it succeeded, or the error if it
   * failed.
   */
  @WorkerThread
  Error enableWorkAuthenticator() {
    if (isWorkAccountAuthenticatorEnabled()) {
      Log.i(Constants.TAG, "Work account authenticator is already enabled.");
      // Still attempt to set app restrictions and invoke API to make sure.
    }

    // Workaround to ensure WorkAccountAuthenticatorInitializer won't disable the authenticator
    // accidentally. This uses legacy API to write restrictions to Play Services, for which
    // WorkAccountAuthenticatorInitializer always reads the value from disk rather than from a
    // member variable, to avoid concurrency issues.
    if (devicePolicyManagerHelper.isDeviceOrProfileOwner()) {
      DevicePolicyManager devicePolicyManager =
          (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
      Bundle originalBundle =
          devicePolicyManager.getApplicationRestrictions(
              admin, GoogleApiAvailability.GOOGLE_PLAY_SERVICES_PACKAGE);
      Bundle newBundle = new Bundle(originalBundle);
      newBundle.putBoolean(ENABLE_WORK_ACCOUNTS_RESTRICTION, true);
      devicePolicyManager.setApplicationRestrictions(
          admin, GoogleApiAvailability.GOOGLE_PLAY_SERVICES_PACKAGE, newBundle);
    }

    Log.i(Constants.TAG, "Enabling work authenticator.");
    workAccountApiHelper.enableWorkAuthenticator();

    Log.i(Constants.TAG, "Ensuring work authenticator is enabled.");
    long maxTime = SystemClock.uptimeMillis() + ENABLE_WORK_AUTHENTICATOR_TIMEOUT_MS;
    while (SystemClock.uptimeMillis() < maxTime) {
      if (isWorkAccountAuthenticatorEnabled()) {
        Log.i(Constants.TAG, "Work authenticator enabled.");
        return null;
      }
      SystemClock.sleep(ENABLE_WORK_AUTHENTICATOR_WAIT_INTERVAL_MS);
    }
    return Error.ENABLE_WORK_ACCOUNT_AUTHENTICATOR_TIMEOUT;
  }

  boolean isWorkAccountAuthenticatorEnabled() {
    AccountManager accountManager = AccountManager.get(context);
    for (AuthenticatorDescription description : accountManager.getAuthenticatorTypes()) {
      if (Constants.WORK_ACCOUNT_TYPE.equals(description.type)) {
        return true;
      }
    }
    return false;
  }
}
