/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.accounts.Account;
import android.content.ComponentName;
import android.content.Context;
import android.text.TextUtils;
import android.util.Log;
import androidx.annotation.VisibleForTesting;
import androidx.annotation.WorkerThread;
import com.google.android.apps.work.dpcsupport.WorkAccountAddedCallback.Error;

/**
 * Helper for adding Android for Work Accounts.
 */
class WorkAccountProvisioner {

  private static final String EMPTY_DEVICE_HINT = "";

  private final Context context;
  private final ComponentName admin;
  private final DevicePolicyManagerHelper devicePolicyManagerHelper;

  private WorkAccountAddedCallback callback;
  private final WorkAccountApiHelper workAccountApiHelper;

  WorkAccountProvisioner(Context context, ComponentName admin) {
    this(context, admin, new WorkAccountApiHelper(context));
  }

  @VisibleForTesting WorkAccountProvisioner(
      Context context, ComponentName admin, WorkAccountApiHelper workAccountApiHelper) {
    this.context = context;
    this.admin = admin;
    this.workAccountApiHelper = workAccountApiHelper;
    devicePolicyManagerHelper = new DevicePolicyManagerHelper(context);
  }

  /**
   * Adds an Android for Work Account to the device using the specified token.
   *
   * @see AndroidForWorkAccountSupport#addAndroidForWorkAccount
   */
  @WorkerThread
  void addWorkAccount(
      String token, int minPlayServicesVersion, final WorkAccountAddedCallback callback) {
    this.callback = callback;

    if (TextUtils.isEmpty(token)) {
      Log.e(Constants.TAG, "Token may not be empty");
      fail(Error.EMPTY_TOKEN);
      return;
    }

    PackageVersionUtils versionUtils = new PackageVersionUtils(context);

    if (!versionUtils.isPlayStoreInstalled()) {
      Log.e(Constants.TAG, "Play Store not installed.");
      fail(Error.FAILED_PRECONDITION, new IllegalStateException("Play Store not installed."));
      return;
    }
    if (!versionUtils.isPlayServicesInstalled()) {
      Log.e(Constants.TAG, "Play Services not installed.");
      fail(Error.FAILED_PRECONDITION, new IllegalStateException("Play Services not installed."));
      return;
    }
    if (!versionUtils.isPlayStoreVersionCompatible()) {
      Log.e(
          Constants.TAG,
          "Play Store not up to date. "
              + "Call ensureWorkingEnvironment before adding an account.");
      fail(Error.FAILED_PRECONDITION, new IllegalStateException("Play Store not up to date."));
      return;
    }
    if (!versionUtils.isPlayServicesVersionCompatible(minPlayServicesVersion)) {
      Log.e(
          Constants.TAG,
          "Play Services not up to date. "
              + "Call ensureWorkingEnvironment before adding an account.");
      fail(Error.FAILED_PRECONDITION, new IllegalStateException("Play Services not up to date."));
      return;
    }

    if (!new WorkAccountAuthenticatorEnabler(
            context, admin, devicePolicyManagerHelper, workAccountApiHelper)
        .isWorkAccountAuthenticatorEnabled()) {
      Log.e(
          Constants.TAG,
          "Work account authenticator has not been enabled. "
              + "Call ensureWorkingEnvironment before adding an account.");
      fail(
          Error.FAILED_PRECONDITION,
          new IllegalStateException("Work account authenticator has not been enabled."));
      return;
    }

    Account addedAccount;
    try {
      addedAccount = workAccountApiHelper.addWorkAccount(token);
    } catch (Exception e) {
      fail(Error.EXCEPTION_ADDING_ACCOUNT, e);
      return;
    }

    if (addedAccount != null) {
      Log.i(Constants.TAG, "Work account added.");
      succeed(addedAccount);
    } else {
      fail(Error.EXCEPTION_ADDING_ACCOUNT);
    }
  }

  private void succeed(final Account account) {
    callback.onAccountReady(account, EMPTY_DEVICE_HINT);
  }

  private void fail(final Error error) {
    callback.onFailure(error);
  }

  private void fail(final Error error, final Exception e) {
    callback.onFailure(error, e);
  }
}
